/******************************************************************************

$Author$
  
$Modtime$
$Revision$

Description: Implementation of class "CBitArray"
             (MFC class CArray "specialized" for bit array)

             The implementation of the private member functions "Copylr",
             "Copyrl", and "X" is derived from "The Standard C++ Library",
             copyright  1995 by P.J. Plauger, published by Prentice-Hall, and
             are used with permission.

$Log$

******************************************************************************/

#include "stdafx.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/*** Definition of class "CBitArray" *****************************************/

/*** Public member functions *************************************************/

/*** Adds a new element to the end of an array, growing the array by 1. ******/
int CBitArray::Add(bool bNewElement)
{
  SetAtGrow(m_nSize, bNewElement);
  return m_nSize - 1;
}

/*** Adds the contents of another array to the end of this array. ************/
int CBitArray:: Append(const CBitArray& src)
{
  if (src.m_nSize > 0)
  {
    if (m_nSize + src.m_nSize > m_nTAllocated * T_BIT)
      SetSize(m_nSize + src.m_nSize);

    Copylr(src, m_nSize, 0, m_nSize + src.m_nSize);
    int nOldSize = m_nSize;
    m_nSize += src.m_nSize;
    return nOldSize;
  }
  else
    return -1;
}

/*** Copies the elements of another array to this array. *********************/
void CBitArray::Copy(const CBitArray& src)
{
  if (&src != this)
  {
    if (src.m_nSize > m_nTAllocated * T_BIT) SetSize(src.m_nSize);

    Copylr(src, 0, 0, src.m_nSize);
    m_nSize = src.m_nSize;
  }
}

/*** Frees any extra memory that was allocated while the array was grown. ****/
/*** This function has no effect on the size or upper bound of the array. ****/
void CBitArray::FreeExtra()
{
  int nTNeeded = (m_nSize + T_BIT-1) / T_BIT;

  if (nTNeeded < m_nTAllocated)
  {
    m_pData       = static_cast<T*>(realloc(m_pData, nTNeeded * sizeof(T)));
    m_nTAllocated = nTNeeded;
  }
}

/*** Inserts one element (or multiple copies of an element) at a specified ***/
/*** index in an array. In the process, it shifts up (by incrementing the  ***/
/*** index) the existing element at this index, and it shifts up all       ***/
/*** elements above it.                                                    ***/
void CBitArray::InsertAt(int nIndex, bool bNewElement, int nCount)
{
  if (nCount > 0)
  {
    int nSize = __max(m_nSize, nIndex);

    if (nSize + nCount > m_nTAllocated * T_BIT) SetSize(nSize + nCount);
    if (nIndex < m_nSize)
      Copyrl(*this, nIndex + nCount, nIndex, m_nSize + nCount);

    const int nIndexBehindEnd    = nIndex + nCount;
    const int nBitsAtEnd         = nIndexBehindEnd % T_BIT;
    const int nIndexTBeforeStart = nIndex == 0 ? -1 : (nIndex - 1) / T_BIT;
    const int nIndexTEnd         = (nIndexBehindEnd - 1) / T_BIT;
    const T   src                = bNewElement ? ~static_cast<T>(0) : 0;
          T   mask               = 0;
          T   save               = 0;

    if (nBitsAtEnd != 0)
    {
      mask = ~static_cast<T>(0) >> nBitsAtEnd;
      save = m_pData[nIndexTEnd];
    }

    nIndex %= T_BIT;
    if (nIndex == 0)
      // set full Ts
      for (int i = nIndexTBeforeStart; ++i <= nIndexTEnd;) m_pData[i] = src;
    else
    {
      // set partial T at start shifted by nDstIndex % T_BIT bits
      m_pData[nIndexTBeforeStart] =
        m_pData[nIndexTBeforeStart] & ~static_cast<T>(0) << (T_BIT - nIndex) |
        src >> nIndex;

      // set full Ts
      for (int i = nIndexTBeforeStart; ++i <= nIndexTEnd;) m_pData[i] = src;
    }

    // merge set partial T and partial original T at end
    if (nBitsAtEnd != 0)
      m_pData[nIndexTEnd] = m_pData[nIndexTEnd] & ~mask | save & mask;

    m_nSize = nSize + nCount;
  }
}

/*** Inserts all the elements from another CBitArray, starting at the ********/
/*** nStartIndex position.                                            ********/
void CBitArray::InsertAt(int nStartIndex, CBitArray* pNewArray)
{
  if (pNewArray->m_nSize > 0)
  {
    int nSize = __max(m_nSize, nStartIndex+1);

    if (nSize + pNewArray->m_nSize > m_nTAllocated * T_BIT)
      SetSize(nSize + pNewArray->m_nSize);
    if (nStartIndex+1 < m_nSize)
      Copyrl(*this, nStartIndex + pNewArray->m_nSize, nStartIndex,
             m_nSize + pNewArray->m_nSize);
    Copylr(*pNewArray, nStartIndex, 0, nStartIndex + pNewArray->m_nSize);

    m_nSize = nSize + pNewArray->m_nSize;
  }
}

/*** Removes all the elements from this array. If the array is already *******/
/*** empty, the function still works.                                  *******/
void CBitArray::RemoveAll()
{
  m_nSize       = 0;
  m_nTAllocated = 0;
  free(m_pData);
  m_pData = 0;
}

/*** Sets the new array element at the specified index. SetAt will not *******/
/*** cause the array to grow.                                          *******/
void CBitArray::SetAt(int nIndex, bool bNewElement)
{
  ASSERT(nIndex >= 0 && nIndex < m_nTAllocated * T_BIT);

  if (bNewElement)
    m_pData[nIndex / T_BIT] |=
      static_cast<T>(1) << (T_BIT - 1 - nIndex % T_BIT);
  else
    m_pData[nIndex / T_BIT] &=
      ~(static_cast<T>(1) << (T_BIT - 1 - nIndex % T_BIT));

  if (nIndex >= m_nSize) m_nSize = nIndex + 1;
}

/*** Sets the array element at the specified index. The array grows      *****/
/*** automatically if necessary (that is, the upper bound is adjusted to *****/
/*** accomodate the new element).                                        *****/
void CBitArray::SetAtGrow(int nIndex, bool bNewElement)
{
  if (nIndex >= m_nTAllocated * T_BIT) SetSize(nIndex + 1);
  SetAt(nIndex, bNewElement);
}

/*** Establishes the size of an ampty or existing array; allocates memory ****/
/*** if necessary. If the new size is smaller than the old size, then the ****/
/*** array is truncated and all unused memory is released.                ****/
void CBitArray::SetSize(int nNewSize, int nGrowBy)
{
  int nGrowByT;

  if (nGrowBy == -1)
    nGrowByT = DefaultGrowByT;
  else
  {
    if (nGrowBy == 0) nGrowBy = 1;
    nGrowByT = (nGrowBy + T_BIT-1) / T_BIT;
  }

  int nTNeeded = (nNewSize + T_BIT-1) / T_BIT;

  if (nTNeeded != m_nTAllocated)
  {
    nTNeeded  = (nTNeeded + nGrowByT-1) / nGrowByT;
    nTNeeded *= nGrowByT;

    if (nTNeeded != m_nTAllocated)
    {
      m_pData       = static_cast<T*>(realloc(m_pData, nTNeeded * sizeof(T)));
      m_nTAllocated = nTNeeded;
      if (m_nSize > m_nTAllocated * T_BIT) m_nSize = m_nTAllocated * T_BIT;
    }
  }
}

/*** Removes one or more elements starting at a specified index in an ********/
/*** array. In the process, it shifts down all the elements above the ********/
/*** removed element(s).                                              ********/
void CBitArray::RemoveAt(int nIndex, int nCount)
{
  ASSERT(nIndex >= 0 && nIndex < m_nSize);

  if (nCount > 0)
  {
    nCount = __min(nCount, m_nSize - nIndex);
    if (nCount > 0)
    {
      Copylr(*this, nIndex, nIndex + nCount, m_nSize);

      m_nSize -= nCount;
    }
  }
}

/*** Private member functions ************************************************/

/*** Copy a bit array from left to right *************************************/
void CBitArray::Copylr(const CBitArray& a, int nDstIndexStart, int nSrcIndex,
                       int nDstIndexBehindEnd)
{
  if (nDstIndexBehindEnd > 0)
  {
    const int nBitsAtEnd            = nDstIndexBehindEnd % T_BIT;
    const int nDstIndexTBeforeStart =
      nDstIndexStart == 0 ? -1 : (nDstIndexStart - 1) / T_BIT;
    const int nSrcIndexT            = nSrcIndex / T_BIT;
    const int nDstIndexTEnd         = (nDstIndexBehindEnd - 1) / T_BIT;
    const int nOff                  = nSrcIndexT - nDstIndexTBeforeStart - 1;
          T   mask                  = 0;
          T   save                  = 0;

    if (nBitsAtEnd != 0)
    {
      mask = ~static_cast<T>(0) >> nBitsAtEnd;
      save = m_pData[nDstIndexTEnd];
    }

    nSrcIndex      %= T_BIT;
    nDstIndexStart %= T_BIT;
    if (nDstIndexStart == 0)
      // copy full Ts
      for (int i = nDstIndexTBeforeStart; ++i <= nDstIndexTEnd;)
        m_pData[i] = a.X(i + nOff, nSrcIndex);
    else
    {
      // copy partial T at start shifted by nDstIndex % T_BIT bits
      m_pData[nDstIndexTBeforeStart] =
        m_pData[nDstIndexTBeforeStart] &
        ~static_cast<T>(0) << (T_BIT - nDstIndexStart) |
        a.X(nSrcIndexT, nSrcIndex) >> nDstIndexStart;

      // copy full Ts shifted by nDstIndex % T_BIT bits
      for (int i = nDstIndexTBeforeStart; ++i <= nDstIndexTEnd;)
        m_pData[i] = a.X(i + nOff, nSrcIndex) << (T_BIT - nDstIndexStart) |
                     a.X(i + nOff + 1, nSrcIndex) >> nDstIndexStart;
    }

    // merge copied partial T and partial original T at end
    if (nBitsAtEnd != 0)
      m_pData[nDstIndexTEnd] = m_pData[nDstIndexTEnd] & ~mask | save & mask;
  }
}

/*** Copy a bit array from right to left *************************************/
void CBitArray::Copyrl(const CBitArray& a, int nDstIndexStart, int nSrcIndex,
                       int nDstIndexBehindEnd)
{
  if (nDstIndexBehindEnd > 0)
  {
    const int nBitsAtEnd            = nDstIndexBehindEnd % T_BIT;
    const int nDstIndexTBeforeStart =
      nDstIndexStart == 0 ? -1 : (nDstIndexStart - 1) / T_BIT;
    const int nSrcIndexT            = nSrcIndex / T_BIT;
    const int nDstIndexTEnd         = (nDstIndexBehindEnd - 1) / T_BIT;
    const int nOff                  = nSrcIndexT - nDstIndexTBeforeStart - 1;
          T   mask                  = 0;
          T   save                  = 0;

    if (nBitsAtEnd != 0)
    {
      mask = ~static_cast<T>(0) >> nBitsAtEnd;
      save = m_pData[nDstIndexTEnd];
    }

    nSrcIndex      %= T_BIT;
    nDstIndexStart %= T_BIT;
    if (nDstIndexStart == 0)
      // copy full Ts
      for (int i = nDstIndexTEnd; i > nDstIndexTBeforeStart; i--)
        m_pData[i] = a.X(i + nOff, nSrcIndex);
    else
    {
      // copy full Ts shifted by nDstIndex % T_BIT bits
      for (int i = nDstIndexTEnd; i > nDstIndexTBeforeStart; i--)
        m_pData[i] = a.X(i + nOff, nSrcIndex) << (T_BIT - nDstIndexStart) |
                     a.X(i + nOff + 1, nSrcIndex) >> nDstIndexStart;

      // copy partial T at start shifted by nDstIndex % T_BIT bits
      m_pData[nDstIndexTBeforeStart] =
        m_pData[nDstIndexTBeforeStart] &
        ~static_cast<T>(0) << (T_BIT - nDstIndexStart) |
        a.X(nSrcIndexT, nSrcIndex) >> nDstIndexStart;
    }

    // merge copied partial T and partial original T at end
    if (nBitsAtEnd != 0)
      m_pData[nDstIndexTEnd] = m_pData[nDstIndexTEnd] & ~mask | save & mask;
  }
}
